/*
 * Copyright (C) 2001-2003 by egnite Software GmbH. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY EGNITE SOFTWARE GMBH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL EGNITE
 * SOFTWARE GMBH OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.ethernut.de/
 *
 * The 9-bit communication had been contributed by Brett Abbott, 
 * Digital Telemetry Limited.
 *
 * Dave Smart contributed the synchronous mode support.
 */

/*
 * $Log$
 */

#include <sys/atom.h>
#include <sys/event.h>
#include <sys/timer.h>

#include <dev/irqreg.h>

#include "usartavr.h"

/*
 * Local function prototypes.
 */
static u_long AvrUsart0GetSpeed(void);
static int AvrUsart0SetSpeed(u_long rate);
static u_char AvrUsart0GetDataBits(void);
static int AvrUsart0SetDataBits(u_char bits);
static u_char AvrUsart0GetParity(void);
static int AvrUsart0SetParity(u_char mode);
static u_char AvrUsart0GetStopBits(void);
static int AvrUsart0SetStopBits(u_char bits);
static u_long AvrUsart0GetFlowControl(void);
static int AvrUsart0SetFlowControl(u_long flags);
static u_long AvrUsart0GetStatus(void);
static int AvrUsart0SetStatus(u_long flags);
static u_char AvrUsart0GetClockMode(void);
static int AvrUsart0SetClockMode(u_char mode);
static void AvrUsart0TxStart(void);
static void AvrUsart0RxStart(void);
static int AvrUsart0Init(void);
static int AvrUsart0Deinit(void);


/*!
 * \brief USART0 device control block structure.
 */
static USARTDCB dcb_usart0 = {
    0,                          /* dcb_modeflags */
    0,                          /* dcb_statusflags */
    0,                          /* dcb_rtimeout */
    0,                          /* dcb_wtimeout */
    {0, 0, 0, 0, 0, 0, 0, 0},   /* dcb_tx_rbf */
    {0, 0, 0, 0, 0, 0, 0, 0},   /* dcb_rx_rbf */
    0,                          /* dbc_last_eol */
    AvrUsart0Init,              /* dcb_init */
    AvrUsart0Deinit,            /* dcb_deinit */
    AvrUsart0TxStart,           /* dcb_tx_start */
    AvrUsart0RxStart,           /* dcb_rx_start */
    AvrUsart0SetFlowControl,    /* dcb_set_flow_control */
    AvrUsart0GetFlowControl,    /* dcb_get_flow_control */
    AvrUsart0SetSpeed,          /* dcb_set_speed */
    AvrUsart0GetSpeed,          /* dcb_get_speed */
    AvrUsart0SetDataBits,       /* dcb_set_data_bits */
    AvrUsart0GetDataBits,       /* dcb_get_data_bits */
    AvrUsart0SetParity,         /* dcb_set_parity */
    AvrUsart0GetParity,         /* dcb_get_parity */
    AvrUsart0SetStopBits,       /* dcb_set_stop_bits */
    AvrUsart0GetStopBits,       /* dcb_get_stop_bits */
    AvrUsart0SetStatus,         /* dcb_set_status */
    AvrUsart0GetStatus,         /* dcb_get_status */
    AvrUsart0SetClockMode,      /* dcb_set_clock_mode */
    AvrUsart0GetClockMode,      /* dcb_get_clock_mode */
};

/*!
 * \brief USART0 device information structure.
 *
 * A pointer to this structure must be passed to NutRegisterDevice() 
 * to bind this device driver to the Nut/OS kernel.
 */
NUTDEVICE devUsart0 = {
    0,                          /* Pointer to next device, dev_next. */
    {'u', 's', 'a', 'r', 't', '0', 0, 0, 0},    /* Unique device name, dev_name. */
    IFTYP_CHAR,                 /* Type of device, dev_type. */
    0,                          /* Base address, dev_base (not used). */
    0,                          /* First interrupt number, dev_irq (not used). */
    0,                          /* Interface control block, dev_icb (not used). */
    &dcb_usart0,                /* Driver control block, dev_dcb. */
    UsartInit,                  /* Driver initialization routine, dev_init. */
    UsartIOCtl,                 /* Driver specific control function, dev_ioctl. */
    UsartRead,                  /* Read from device, dev_read. */
    UsartWrite,                 /* Write to device, dev_write. */
    UsartWrite_P,               /* Write data from program space to device, dev_write_P. */
    UsartOpen,                  /* Open a device or file, dev_open. */
    UsartClose,                 /* Close a device or file, dev_close. */
    0                           /* Request file size, dev_size. */
};

/* \brief ASCII code for software flow control, starts transmitter. */
#define ASCII_XON   0x11
/* \brief ASCII code for software flow control, stops transmitter. */
#define ASCII_XOFF  0x13

/* \brief XON transmit pending flag. */
#define XON_PENDING     0x10
/* \brief XOFF transmit pending flag. */
#define XOFF_PENDING    0x20
/* \brief XOFF sent flag. */
#define XOFF_SENT       0x40
/* \brief XOFF received flag. */
#define XOFF_RCVD       0x80


/*!
 * \brief Receiver error flags.
 */
static ureg_t rx_errors;

/*!
 * \brief Enables software flow control if not equal zero.
 */
static ureg_t flow_control;

/*!
 * \brief Transmit address frame, if not zero.
 */
static ureg_t tx_aframe;


#ifdef UART0_HDX_BIT
/*!
 * \brief Enables half duplex control if not equal zero.
 *
 * This variable exists only if the hardware configuration defines a 
 * port bit to switch between receive and transmit mode.
 */
static ureg_t hdx_control;
#endif

/*
 * 9-pin DTE male connector:
 * 1 - DCD input (not supported)
 * 2 - RD input
 * 3 - TD output
 * 4 - DTR output (not supported)
 * 5 - signal ground
 * 6 - DSR input (not supported)
 * 7 - RTS output
 * 8 - CTS input
 * 9 - RI input (not supported)
 *
 * 9-pin DCE female connector:
 * 1 - DCD output (not supported)
 * 2 - RD output
 * 3 - TD input
 * 4 - DTR input (not supported)
 * 5 - signal ground
 * 6 - DSR output (not supported)
 * 7 - RTS input
 * 8 - CTS output
 * 9 - RI output (not supported)
 */

#ifdef UART0_RTS_BIT
/*!
 * \brief Enables RTS control if not equal zero.
 *
 * This variable exists only if the hardware configuration defines a 
 * port bit to control the RTS signal.
 */
static ureg_t rts_control;
#endif

#ifdef UART0_CTS_BIT
/*!
 * \brief Enables CTS sense if not equal zero.
 *
 * This variable exists only if the hardware configuration defines a 
 * port bit to sense the CTS signal.
 */
static ureg_t cts_sense;
#endif

#ifdef UART0_CTS_BIT
/*!
 * \brief USART0 CTS sense interrupt handler.
 *
 * This interrupt routine is called when the CTS line level is low.
 * Typical line drivers negate the signal, thus driving our port
 * low when CTS is active.
 *
 * This routine exists only if the hardware configuration defines a 
 * port bit to sense the CTS signal.
 */
static void AvrUsart0Cts(void *arg)
{
    /* Enable transmit interrupt. */
    sbi(UCSR0B, UDRIE);
    /* Disable CTS sense interrupt. */
    cbi(EIMSK, UART0_CTS_BIT);
}
#endif

#ifdef UART0_HDX_BIT
/*
 * \brief USART0 transmit complete interrupt handler.
 *
 * Used with half duplex communication to switch from tranmit to receive
 * mode after the last character has been transmitted.
 *
 * This routine exists only if the hardware configuration defines a 
 * port bit to switch between receive and transmit mode.
 *
 * \param arg Pointer to the transmitter ring buffer.
 */
static void AvrUsart0TxComplete(void *arg)
{
    register RINGBUF *rbf = (RINGBUF *) arg;

    /*
     * Check if half duplex mode has been enabled and if all characters
     * had been sent out.
     */
    if (hdx_control && rbf->rbf_cnt == 0) {
        /* Switch to receiver mode. */
        cbi(UART0_HDX_PORT, UART0_HDX_BIT);
    }
}
#endif

/*
 * \brief USART0 transmit data register empty interrupt handler.
 *
 * \param arg Pointer to the transmitter ring buffer.
 */
static void AvrUsart0TxEmpty(void *arg)
{
    register RINGBUF *rbf = (RINGBUF *) arg;
    register u_char *cp = rbf->rbf_tail;

    /*
     * Process pending software flow controls first.
     */
    if (flow_control & (XON_PENDING | XOFF_PENDING)) {
        if (flow_control & XON_PENDING) {
            outb(UDR0, ASCII_XOFF);
            flow_control |= XOFF_SENT;
        } else {
            outb(UDR0, ASCII_XON);
            flow_control &= ~XOFF_SENT;
        }
        flow_control &= ~(XON_PENDING | XOFF_PENDING);
        return;
    }

    if (rbf->rbf_cnt--) {

#ifdef UART0_CTS_BIT
        /* 
         * If CTS has been disabled, we disable the transmit interrupts
         * and return without sending anything.
         */
        if (cts_sense && bit_is_set(UART0_CTS_PIN, UART0_CTS_BIT)) {
            cbi(UCSR0B, UDRIE);
            sbi(EIMSK, UART0_CTS_BIT);
            return;
        }
#endif

        /*
         * The data sheet doesn't exactly tell us, if this bit is retained 
         * or cleared after the character has been sent out. So we do it
         * the save way.
         */
        if(tx_aframe) {
            sbi(UCSR0B, TXB8);
        }
        else {
            cbi(UCSR0B, TXB8);
        }

        /*
         * Start transmission of the next character.
         */
        outb(UDR0, *cp);

        /*
         * Wrap around the buffer pointer if we reached its end.
         */
        if (++cp == rbf->rbf_last) {
            cp = rbf->rbf_start;
        }
        rbf->rbf_tail = cp;
        if (rbf->rbf_cnt == rbf->rbf_lwm) {
            NutEventPostAsync(&rbf->rbf_que);
        }
    }

    /*
     * Nothing left to transmit, disable interrupt.
     */
    else {
        cbi(UCSR0B, UDRIE);
        rbf->rbf_cnt = 0;
        NutEventPostAsync(&rbf->rbf_que);
    }
}

/*
 * \brief USART0 receive complete interrupt handler.
 *
 *
 * \param arg Pointer to the receiver ring buffer.
 */
void AvrUsart0RxComplete(void *arg)
{
    register RINGBUF *rbf = (RINGBUF *) arg;
    register size_t cnt;
    register u_char ch;

    /*
     * We read the received character as early as possible to avoid overflows
     * caused by interrupt latency. However, reading the error flags must come 
     * first, because reading the ATmega128 data register clears the status.
     */
    rx_errors |= inb(UCSR0A);
    ch = inb(UDR0);

    /*
     * Handle software handshake. We have to do this before checking the
     * buffer, because flow control must work in write-only mode, where
     * there is no receive buffer.
     */
    if (flow_control) {
        /* XOFF character disables transmit interrupts. */
        if (ch == ASCII_XOFF) {
            cbi(UCSR0B, UDRIE);
            flow_control |= XOFF_RCVD;
            return;
        }
        /* XON enables transmit interrupts. */
        else if (ch == ASCII_XON) {
            sbi(UCSR0B, UDRIE);
            flow_control &= ~XOFF_RCVD;
            return;
        }
    }

    /*
     * Check buffer overflow.
     */
    cnt = rbf->rbf_cnt;
    if (cnt >= rbf->rbf_siz) {
        rx_errors |= _BV(DOR);
        return;
    }

    /* Wake up waiting threads if this is the first byte in the buffer. */
    if (cnt++ == 0) {
        NutEventPostAsync(&rbf->rbf_que);
    }

    /*
     * Check the high watermark for software handshake. If the number of 
     * buffered bytes is above this mark, then send XOFF.
     */
    else if (flow_control && (flow_control & XOFF_SENT) == 0 && cnt >= rbf->rbf_hwm) {
        if (inb(UCSR0A) & _BV(UDRE)) {
            outb(UDR0, ASCII_XOFF);
            flow_control |= XOFF_SENT;
            flow_control &= ~XOFF_PENDING;
        } else {
            flow_control |= XOFF_PENDING;
        }
    }
#ifdef UART0_RTS_BIT
    /*
     * Check the high watermark for hardware handshake. If the number of 
     * buffered bytes is above this mark, then disable RTS.
     */
    else if (rts_control && cnt >= rbf->rbf_hwm) {
        sbi(UART0_RTS_PORT, UART0_RTS_BIT);
    }
#endif

    /* 
     * Store the character and increment and the ring buffer pointer. 
     */
    *rbf->rbf_head++ = ch;
    if (rbf->rbf_head == rbf->rbf_last) {
        rbf->rbf_head = rbf->rbf_start;
    }

    /* Update the ring buffer counter. */
    rbf->rbf_cnt = cnt;
}


/*!
 * \brief Carefully enable USART hardware functions.
 *
 * Always enabale transmitter and receiver, even on read-only or
 * write-only mode. So we can support software flow control.
 */
static void AvrUsart0Enable(void)
{
    NutEnterCritical();

    outb(UCSR0B, _BV(RXCIE) | _BV(UDRIE) | _BV(RXEN) | _BV(TXEN));

#ifdef UART0_HDX_BIT
    if (hdx_control) {
        /* Enable transmit complete interrupt. */
        sbi(UCSR0B, TXCIE);
    }
#endif

    NutExitCritical();
}

/*!
 * \brief Carefully disable USART hardware functions.
 */
static void AvrUsart0Disable(void)
{
    /*
     * Disable USART interrupts.
     */
    NutEnterCritical();
    cbi(UCSR0B, RXCIE);
    cbi(UCSR0B, TXCIE);
    cbi(UCSR0B, UDRIE);
    NutExitCritical();

    /*
     * Allow incoming or outgoing character to finish.
     */
    NutDelay(10);

    /*
     * Disable USART transmit and receive.
     */
    cbi(UCSR0B, RXEN);
    cbi(UCSR0B, TXEN);
}

/*!
 * \brief Query the USART hardware for the selected speed.
 *
 * This function is called by ioctl function of the upper level USART 
 * driver through the USARTDCB jump table.
 *
 * \return The currently selected baudrate.
 */
static u_long AvrUsart0GetSpeed(void)
{
    u_long fct;
    u_short sv = ((u_short) inb(UBRR0H) << 8) | (u_short) inb(UBRR0L);

    /* Synchronous mode. */
    if (bit_is_set(UCSR0C, UMSEL)) {
        fct = 2UL;
    }

    /* Double rate mode. */
    else if (bit_is_set(UCSR0A, U2X)) {
        fct = 8UL;
    }

    /* Normal mode. */
    else {
        fct = 16UL;
    }
    return NutGetCpuClock() / (fct * ((u_long)sv + 1UL));
}

/*!
 * \brief Set the USART hardware bit rate.
 *
 * This function is called by ioctl function of the upper level USART 
 * driver through the USARTDCB jump table.
 *
 * \param rate Number of bits per second.
 *
 * \return 0 on success, -1 otherwise.
 */
static int AvrUsart0SetSpeed(u_long rate)
{
    u_short sv;

    AvrUsart0Disable();

    /* 
     * Modified Robert Hildebrand's refined calculation.
     */
    if (bit_is_clear(UCSR0C, UMSEL)) {
        if (bit_is_set(UCSR0A, U2X)) {
            rate <<= 2;
        }
        else {
            rate <<= 3;
        }
    }
    sv = (u_short) ((NutGetCpuClock() / rate + 1UL) / 2UL) - 1;

    outb(UBRR0L, (u_char) sv);
    outb(UBRR0H, (u_char) (sv >> 8));
    AvrUsart0Enable();

    return 0;
}

/*!
 * \brief Query the USART hardware for the number of data bits.
 *
 * This function is called by ioctl function of the upper level USART 
 * driver through the USARTDCB jump table.
 *
 * \return The number of data bits set.
 */
static u_char AvrUsart0GetDataBits(void)
{
    if (bit_is_set(UCSR0B, UCSZ2)) {
        return 9;
    }
    if (bit_is_set(UCSR0C, UCSZ1)) {
        if (bit_is_set(UCSR0C, UCSZ0)) {
            return 8;
        } else {
            return 7;
        }
    } else if (bit_is_set(UCSR0C, UCSZ0)) {
        return 6;
    }
    return 5;
}

/*!
 * \brief Set the USART hardware to the number of data bits.
 *
 * This function is called by ioctl function of the upper level USART 
 * driver through the USARTDCB jump table.
 *
 * \return 0 on success, -1 otherwise.
 */
static int AvrUsart0SetDataBits(u_char bits)
{
    AvrUsart0Disable();
    cbi(UCSR0C, UCSZ0);
    cbi(UCSR0C, UCSZ1);
    cbi(UCSR0B, UCSZ2);
    switch (bits) {
    case 6:
        sbi(UCSR0C, UCSZ0);
        break;
    case 9:
        sbi(UCSR0B, UCSZ2);
    case 8:
        sbi(UCSR0C, UCSZ0);
    case 7:
        sbi(UCSR0C, UCSZ1);
        break;
    }
    AvrUsart0Enable();

    /*
     * Verify the result.
     */
    if (AvrUsart0GetDataBits() != bits) {
        return -1;
    }
    return 0;
}

/*!
 * \brief Query the USART hardware for the parity mode.
 *
 * This routine is called by ioctl function of the upper level USART 
 * driver through the USARTDCB jump table.
 *
 * \return Parity mode, either 0 (disabled), 1 (odd) or 2 (even).
 */
static u_char AvrUsart0GetParity(void)
{
    if (bit_is_set(UCSR0C, UPM1)) {
        if (bit_is_set(UCSR0C, UPM0)) {
            return 1;
        } else {
            return 2;
        }
    }
    return 0;
}

/*!
 * \brief Set the USART hardware to the specified parity mode.
 *
 * This routine is called by ioctl function of the upper level USART 
 * driver through the USARTDCB jump table.
 *
 * \param mode 0 (disabled), 1 (odd) or 2 (even)
 *
 * \return 0 on success, -1 otherwise.
 */
static int AvrUsart0SetParity(u_char mode)
{
    AvrUsart0Disable();
    switch (mode) {
    case 0:
        cbi(UCSR0C, UPM0);
        cbi(UCSR0C, UPM1);
        break;
    case 1:
        sbi(UCSR0C, UPM0);
        sbi(UCSR0C, UPM1);
        break;
    case 2:
        cbi(UCSR0C, UPM0);
        sbi(UCSR0C, UPM1);
        break;
    }
    AvrUsart0Enable();

    /*
     * Verify the result.
     */
    if (AvrUsart0GetParity() != mode) {
        return -1;
    }
    return 0;
}

/*!
 * \brief Query the USART hardware for the number of stop bits.
 *
 * This routine is called by ioctl function of the upper level USART 
 * driver through the USARTDCB jump table.
 *
 * \return The number of stop bits set, either 1 or 2.
 */
static u_char AvrUsart0GetStopBits(void)
{
    if (bit_is_set(UCSR0C, USBS)) {
        return 2;
    }
    return 1;
}

/*!
 * \brief Set the USART hardware to the number of stop bits.
 *
 * This routine is called by ioctl function of the upper level USART 
 * driver through the USARTDCB jump table.
 *
 * \return 0 on success, -1 otherwise.
 */
static int AvrUsart0SetStopBits(u_char bits)
{
    AvrUsart0Disable();
    if (bits == 1) {
        cbi(UCSR0C, USBS);
    } else if (bits == 2) {
        sbi(UCSR0C, USBS);
    }
    AvrUsart0Enable();

    /*
     * Verify the result.
     */
    if (AvrUsart0GetStopBits() != bits) {
        return -1;
    }
    return 0;
}

/*!
 * \brief Query the USART hardware status.
 *
 * \return Status flags.
 */
static u_long AvrUsart0GetStatus(void)
{
    u_long rc = 0;

    /*
     * Set receiver error flags.
     */
    if((rx_errors & _BV(FE)) != 0) {
        rc |= UART_FRAMINGERROR;
    }
    if((rx_errors & _BV(DOR)) != 0) {
        rc |= UART_OVERRUNERROR;
    }
    if((rx_errors & _BV(UPE)) != 0) {
        rc |= UART_PARITYERROR;
    }

    /*
     * Determine software handshake status.
     */
    if (flow_control) {
        if(flow_control & XOFF_SENT) {
            rc |= UART_RXDISABLED;
        }
        if(flow_control & XOFF_RCVD) {
            rc |= UART_TXDISABLED;
        }
    }

#ifdef UART0_RTS_BIT
    /*
     * Determine hardware handshake control status.
     */
    if(bit_is_set(UART0_RTS_PORT, UART0_RTS_BIT)) {
        rc |= UART_RTSDISABLED;
        if(rts_control) {
            rc |= UART_RXDISABLED;
        }
    }
    else {
        rc |= UART_RTSENABLED;
    }
#endif

#ifdef UART0_CTS_BIT
    /*
     * Determine hardware handshake sense status.
     */
    if(bit_is_set(UART0_CTS_PIN, UART0_CTS_BIT)) {
        rc |= UART_CTSDISABLED;
        if (cts_sense) {
            rc |= UART_RXDISABLED;
        }
    }
    else {
        rc |= UART_CTSENABLED;
    }
#endif

    /*
     * If transmitter and receiver haven't been detected disabled by any
     * of the checks above, then they are probably enabled.
     */
    if((rc & UART_RXDISABLED) == 0) {
        rc |= UART_RXENABLED;
    }
    if((rc & UART_TXDISABLED) == 0) {
        rc |= UART_TXENABLED;
    }

    /*
     * Process multidrop setting.
     */
    if(tx_aframe) {
        rc |= UART_TXADDRFRAME;
    }
    else {
        rc |= UART_TXNORMFRAME;
    }
    if(bit_is_set(UCSR0A, MPCM)) {
        rc |= UART_RXADDRFRAME;
    }
    else {
        rc |= UART_RXNORMFRAME;
    }

    return rc;
}

/*!
 * \brief Set the USART hardware status.
 *
 * \param flags Status flags.
 *
 * \return 0 on success, -1 otherwise.
 */
static int AvrUsart0SetStatus(u_long flags)
{
    /*
     * Process software handshake control.
     */
    if (flow_control) {
        /*
         * Enabling or disabling the receiver means to behave like 
         * having sent a XON or XOFF character resp.
         */
        if(flags & UART_RXENABLED) {
            flow_control &= ~XOFF_SENT;
        }
        else if(flags & UART_RXDISABLED) {
            flow_control |= XOFF_SENT;
        }

        /*
         * Enabling or disabling the transmitter means to behave like 
         * having received a XON or XOFF character resp.
         */
        if(flags & UART_TXENABLED) {
            flow_control &= ~XOFF_RCVD;
        }
        else if(flags & UART_TXDISABLED) {
            flow_control |= XOFF_RCVD;
        }
    }

#ifdef UART0_RTS_BIT
    /*
     * Process hardware handshake control.
     */
    if(rts_control) {
        if(flags & UART_RXDISABLED) {
            sbi(UART0_RTS_PORT, UART0_RTS_BIT);
        }
        if(flags & UART_RXENABLED) {
            cbi(UART0_RTS_PORT, UART0_RTS_BIT);
        }
    }
    if(flags & UART_RTSDISABLED) {
        sbi(UART0_RTS_PORT, UART0_RTS_BIT);
    }
    if(flags & UART_RTSENABLED) {
        cbi(UART0_RTS_PORT, UART0_RTS_BIT);
    }
#endif

    /*
     * Process multidrop setting.
     */
    if(flags & UART_TXADDRFRAME) {
        tx_aframe = 1;
    }
    if(flags & UART_TXNORMFRAME) {
        tx_aframe = 0;
    }
    if(flags & UART_RXADDRFRAME) {
        sbi(UCSR0A, MPCM);
    }
    if(flags & UART_RXNORMFRAME) {
        cbi(UCSR0A, MPCM);
    }

    /*
     * Clear UART receive errors.
     */
    if(flags & UART_FRAMINGERROR) {
        rx_errors &= ~_BV(FE);
    }
    if(flags & UART_OVERRUNERROR) {
        rx_errors &= ~_BV(DOR);
    }
    if(flags & UART_PARITYERROR) {
        rx_errors &= ~_BV(UPE);
    }

    /*
     * Verify the result.
     */
    if ((AvrUsart0GetStatus() & ~UART_ERRORS) != flags) {
        return -1;
    }
    return 0;
}

/*!
 * \brief Query the USART hardware for synchronous mode.
 *
 * This function is called by ioctl function of the upper level USART 
 * driver through the USARTDCB jump table.
 *
 * \return Or-ed combination of USART_SYNC, USART_MASTER, USART_NCLOCK 
 *         and USART_HIGHSPEED.
 */
static u_char AvrUsart0GetClockMode(void)
{
    u_char rc = 0;

    if (bit_is_set(UCSR0C, UMSEL)) {
        rc |= USART_SYNC;
        if (bit_is_set(DDRE, 2)) {
            rc |= USART_MASTER;
        }
        if (bit_is_set(UCSR0C, UCPOL)) {
            rc |= USART_NCLOCK;
        }
    }
    else if (bit_is_set(UCSR0A, U2X)) {
        rc |= USART_HIGHSPEED;
    }
    return rc;
}

/*!
 * \brief Set asynchronous or synchronous mode.
 *
 * This function is called by ioctl function of the upper level USART 
 * driver through the USARTDCB jump table.
 *
 * \param mode Must be an or-ed combination of USART_SYNC, USART_MASTER,
 *             USART_NCLOCK and USART_HIGHSPEED. 
 *
 * \return 0 on success, -1 otherwise.
 */
static int AvrUsart0SetClockMode(u_char mode)
{
    AvrUsart0Disable();

    /*
     * Handle synchronous mode.
     */
    if(mode & USART_SYNC) {
        if(mode & USART_MASTER) {
            /* Enable master mode. */
            sbi(DDRE, 2);
        }
        else {
            /* Disable master mode. */
            cbi(DDRE, 2);
        }
        if(mode & USART_NCLOCK) {
            /* Enable negated clock. */
            sbi(UCSR0C, UCPOL);
        }
        else {
            /* Disable negated clock. */
            cbi(UCSR0C, UCPOL);
        }
        /* Disable high speed. */
        cbi(UCSR0A, U2X);
        /* Enable synchronous mode. */
        sbi(UCSR0C, UMSEL);
    }

    /*
     * Handle asynchronous mode.
     */
    else {
        if(mode & USART_HIGHSPEED) {
            /* Enable high speed. */
            sbi(UCSR0A, U2X);
        }
        else {
            /* Disable high speed. */
            cbi(UCSR0A, U2X);
        }
        /* Disable negated clock. */
        cbi(UCSR0C, UCPOL);
        /* Disable synchronous mode. */
        cbi(UCSR0C, UMSEL);
    }
    AvrUsart0Enable();

    /*
     * Verify the result.
     */
    if (AvrUsart0GetClockMode() != mode) {
        return -1;
    }
    return 0;
}

/*!
 * \brief Query flow control mode.
 *
 * This routine is called by ioctl function of the upper level USART 
 * driver through the USARTDCB jump table.
 *
 * \return See UsartIOCtl().
 */
static u_long AvrUsart0GetFlowControl(void)
{
    u_long rc = 0;

    if (flow_control) {
        rc |= USART_MF_XONXOFF;
    } else {
        rc &= ~USART_MF_XONXOFF;
    }

#ifdef UART0_RTS_BIT
    if (rts_control) {
        rc |= USART_MF_RTSCONTROL;
    } else {
        rc &= ~USART_MF_RTSCONTROL;
    }
#endif

#ifdef UART0_CTS_BIT
    if (cts_sense) {
        rc |= USART_MF_CTSSENSE;
    } else {
        rc &= ~USART_MF_CTSSENSE;
    }
#endif

#ifdef UART0_HDX_BIT
    if (hdx_control) {
        rc |= USART_MF_HALFDUPLEX;
    } else {
        rc &= ~USART_MF_HALFDUPLEX;
    }
#endif

    return rc;
}

/*!
 * \brief Set flow control mode.
 *
 * This function is called by ioctl function of the upper level USART 
 * driver through the USARTDCB jump table.
 *
 * \param flags See UsartIOCtl().
 *
 * \return 0 on success, -1 otherwise.
 */
static int AvrUsart0SetFlowControl(u_long flags)
{
    if (flags & USART_MF_XONXOFF) {
        flow_control = 1;
    } else {
        flow_control = 0;
    }

#ifdef UART0_RTS_BIT
    if (flags & USART_MF_RTSCONTROL) {
        sbi(UART0_RTS_PORT, UART0_RTS_BIT);
        sbi(UART0_RTS_DDR, UART0_RTS_BIT);
        rts_control = 1;
    } else if (rts_control) {
        rts_control = 0;
        cbi(UART0_RTS_DDR, UART0_RTS_BIT);
    }
#endif

#ifdef UART0_CTS_BIT
    if (flags & USART_MF_CTSSENSE) {
        /* Register CTS sense interrupt. */
        if (NutRegisterIrqHandler(&UART0_CTS_SIGNAL, AvrUsart0Cts, 0)) {
            return -1;
        }
        sbi(UART0_CTS_PORT, UART0_CTS_BIT);
        cbi(UART0_CTS_DDR, UART0_CTS_BIT);
        cts_sense = 1;
    } else if (cts_sense) {
        cts_sense = 0;
        /* Deregister CTS sense interrupt. */
        NutRegisterIrqHandler(&UART0_CTS_SIGNAL, 0, 0);
        cbi(UART0_CTS_DDR, UART0_CTS_BIT);
    }
#endif

#ifdef UART0_HDX_BIT
    if (flags & USART_MF_HALFDUPLEX) {
        /* Register transmit complete interrupt. */
        if (NutRegisterIrqHandler(&sig_UART0_TRANS, AvrUsart0TxComplete, &dcb_usart0.dcb_rx_rbf)) {
            return -1;
        }
        /* Initially enable the receiver. */
        cbi(UART0_HDX_PORT, UART0_HDX_BIT);
        sbi(UART0_HDX_DDR, UART0_HDX_BIT);
        hdx_control = 1;
    } else if (hdx_control) {
        hdx_control = 0;
        /* Deregister transmit complete interrupt. */
        NutRegisterIrqHandler(&sig_UART0_TRANS, 0, 0);
        cbi(UART0_HDX_DDR, UART0_HDX_BIT);
    }
#endif

    /*
     * Verify the result.
     */
    if (AvrUsart0GetFlowControl() != flags) {
        return -1;
    }
    return 0;
}

/*!
 * \brief Start the USART transmitter hardware.
 *
 * The upper level USART driver will call this function through the 
 * USARTDCB jump table each time it added one or more bytes to the 
 * transmit buffer.
 */
static void AvrUsart0TxStart(void)
{
#ifdef UART0_HDX_BIT
    if (hdx_control) {
        /* Enable half duplex transmitter. */
        sbi(UART0_RTS_PORT, UART0_RTS_BIT);
    }
#endif
    /* Enable transmit interrupts. */
    sbi(UCSR0B, UDRIE);
}

/*!
 * \brief Start the USART receiver hardware.
 *
 * The upper level USART driver will call this function through the 
 * USARTDCB jump table each time it removed enough bytes from the 
 * receive buffer. Enough means, that the number of bytes left in
 * the buffer is below the low watermark.
 */
static void AvrUsart0RxStart(void)
{
    /*
     * Do any required software flow control.
     */
    if (flow_control && ((flow_control & XOFF_SENT) != 0 || mode == 1)) {
        NutEnterCritical();
        if (inb(UCSR0A) & _BV(UDRE)) {
            outb(UDR0, ASCII_XON);
            flow_control &= ~XON_PENDING;
        } else {
            flow_control |= XON_PENDING;
        }
        flow_control &= (XOFF_SENT | XOFF_PENDING);
        NutExitCritical();
    }

#ifdef UART0_RTS_BIT
    if (rts_control) {
        /* Enable RTS. */
        cbi(UART0_RTS_PORT, UART0_RTS_BIT);
    }
#endif
}

/*
 * \brief Initialize the USART hardware driver.
 *
 * This function is called during device registration by the upper level 
 * USART driver through the USARTDCB jump table.
 *
 * \return 0 on success, -1 otherwise.
 */
static int AvrUsart0Init(void)
{
    /*
     * Register receive and transmit interrupts.
     */
    if (NutRegisterIrqHandler(&sig_UART0_RECV, AvrUsart0RxComplete, &dcb_usart0.dcb_rx_rbf))
        return -1;
    if (NutRegisterIrqHandler(&sig_UART0_DATA, AvrUsart0TxEmpty, &dcb_usart0.dcb_tx_rbf)) {
        NutRegisterIrqHandler(&sig_UART0_RECV, 0, 0);
        return -1;
    }
    return 0;
}

/*
 * \brief Deinitialize the USART hardware driver.
 *
 * This function is called during device deregistration by the upper 
 * level USART driver through the USARTDCB jump table.
 *
 * \return 0 on success, -1 otherwise.
 */
static int AvrUsart0Deinit(void)
{
    /* Deregister receive and transmit interrupts. */
    NutRegisterIrqHandler(&sig_UART0_RECV, 0, 0);
    NutRegisterIrqHandler(&sig_UART0_DATA, 0, 0);

    /*
     * Disabling flow control shouldn't be required here, because it's up
     * to the upper level to do this on the last close or during
     * deregistration.
     */
#ifdef UART0_HDX_BIT
    /* Deregister transmit complete interrupt. */
    if (hdx_control) {
        hdx_control = 0;
        NutRegisterIrqHandler(&sig_UART0_TRANS, 0, 0);
    }
#endif

#ifdef UART0_CTS_BIT
    if (cts_sense) {
        cts_sense = 0;
        cbi(UART0_CTS_DDR, UART0_CTS_BIT);
        /* Deregister CTS sense interrupt. */
        NutRegisterIrqHandler(&UART0_CTS_SIGNAL, 0, 0);
    }
#endif

#ifdef UART0_RTS_BIT
    if (rts_control) {
        rts_control = 0;
        cbi(UART0_RTS_DDR, UART0_RTS_BIT);
    }
#endif

    return 0;
}
